/* -*-C-*-
 ******************************************************************************
 *
 * File:        globals.h  (Formerly globals.h)
 * Description:  Global Variables for Wise Owl
 * Author:       Mark Seaman, OCR Technology
 * Created:      Thu Dec 21 11:38:36 1989
 * Modified:     Thu Jan  4 17:13:00 1990 (Mark Seaman) marks@hpgrlt
 * Language:     C
 * Package:      N/A
 * Status:       Experimental (Do Not Distribute)
 *
 * (c) Copyright 1989, Hewlett-Packard Company.
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
 ** You may obtain a copy of the License at
 ** http://www.apache.org/licenses/LICENSE-2.0
 ** Unless required by applicable law or agreed to in writing, software
 ** distributed under the License is distributed on an "AS IS" BASIS,
 ** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 ** See the License for the specific language governing permissions and
 ** limitations under the License.
 *
 *****************************************************************************/
#ifndef GLOBALS_H
#define GLOBALS_H

#include "unicharset.h"
#include "strngs.h"

#include <cstdio>

#endif
